"""
Spawning and cleaning after given HTTP service processes and Mountebank.
"""

from mountepy.http_service import HttpService, ServiceGroup, wait_for_port
from mountepy.mountebank import Mountebank, ExistingMountebank, HttpStub



"""
Management of Mountebank installation.
Can detect Mountebank installed in the system and also download and setup a standalone distribution
"""

from functools import lru_cache
import logging
import os
import subprocess
import tarfile
import urllib


CACHE_DIR = os.path.expanduser('~/.cache/mountepy')
MB_INSTALL_CHECK_CMD = ['mb', 'help']

_log = logging.getLogger(__name__)  # pylint: disable=invalid-name


@lru_cache(1)
def get_mb_command():
    """Gets the command to start a Mountebank server in the system.
    Ensures that either Mountebank installation or standalone distribution is available.
    Will trigger Mountebank download if necessary.

    Returns:
        list[str]: Command (Popen-compatible) that starts Mountebank on this system.
    """
    if _check_mb_install():
        return ['mb']
    else:
        if not os.path.exists(CACHE_DIR):
            _setup_standalone()
        mb_dir = _get_mb_dir(CACHE_DIR)
        node_path = _get_node_path(mb_dir)
        mb_exe_path = os.path.join(mb_dir, 'mountebank/bin/mb')
        mb_command = [node_path, mb_exe_path]
        _log.debug('Standalone Mountebank command set to %s', mb_command)
        return mb_command


def _setup_standalone():
    """Downloads and sets up a standalone distribution Mountebank.
    Standalone distribution also contains NodeJS.
    """
    _log.info('Setting up a standalone Mountebank.')
    mb_archive_path, _ = urllib.request.urlretrieve(
        'https://s3.amazonaws.com/mountebank/v1.4/mountebank-v1.4.3-linux-x64.tar.gz')
    mb_tar = tarfile.open(mb_archive_path, mode='r:gz')
    mb_tar.extractall(CACHE_DIR)

    os.remove(mb_archive_path)


def _check_mb_install():
    """Checks if Mountebank is normally installed in the system.

    Returns:
        bool: True if Mountebank is installed, False otherwise.
    """
    try:
        subprocess.check_call(
            MB_INSTALL_CHECK_CMD,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL)
        _log.debug('Detected normal Mountebank installation.')
        return True
    except FileNotFoundError:
        _log.debug('No normal Mountebank installation found.')
        return False


def _get_mb_dir(cache_dir):
    try:
        mb_dir = next(path for path in os.listdir(cache_dir) if path.startswith('mountebank-v'))
        return os.path.join(cache_dir, mb_dir)
    except StopIteration:
        raise MBStandaloneBrokenError(
            "No Mountebank distribution found! "
            "It should have been automatically downloaded earlier")


def _get_node_path(mb_dir):
    try:
        node_path = next(path for path in os.listdir(mb_dir) if path.startswith('node-v'))
        return os.path.join(mb_dir, node_path, 'bin/node')
    except StopIteration:
        raise MBStandaloneBrokenError('No NodeJS distribution found in Mountebank distribution! '
                                      'WTF? Distribution dir: ' + mb_dir)


class MBStandaloneBrokenError(Exception):
    """Means that the standalone Mountebank setup is broken.
    Probably someone deleted some files from it.
    """



# -*- coding: utf-8 -*-
import os.path
from setuptools import setup

project_name = 'mountepy'
version = '0.4.1'

setup_dir = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(setup_dir, 'requirements.txt')) as req_file:
    requirements = [lib.split('==')[0] for lib in req_file.readlines()]
with open(os.path.join(setup_dir, 'README.rst')) as readme_file:
    readme = readme_file.read()

setup(
    name=project_name,
    version=version,
    description='Utilities for creating (micro)service tests. Based on Mountebank.',
    long_description=readme,
    author='Michał Bultrowicz',
    author_email='michal.bultrowicz@gmail.com',
    url='https://github.com/butla/mountepy',
    packages=[
        project_name,
    ],
    package_dir={project_name: project_name},
    include_package_data=True,
    install_requires=requirements,
    license="BSD Zero Clause",
    keywords='test http mountebank microservice',
    classifiers=[
        'Programming Language :: Python :: 3.4',
        'Operating System :: POSIX :: Linux',
        'Development Status :: 4 - Beta',
        'License :: Freely Distributable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
    ],
)



"""
Abstractions for management of HTTP service processes.
"""

import atexit
import concurrent.futures
import logging
import os
import signal
import socket
import subprocess
import time

import port_for


def wait_for_port(port, host='localhost', timeout=5.0):
    """Wait until a port starts accepting TCP connections.

    Args:
        port (int): Port number.
        host (str): Host address on which the port should exist.
        timeout (float): In seconds. How long to wait before raising errors.

    Raises:
        TimeoutError: The port isn't accepting connection after time specified in `timeout`.
    """
    start_time = time.perf_counter()
    while True:
        try:
            with socket.create_connection((host, port)):
                break
        except OSError:
            time.sleep(0.01)
            if time.perf_counter() - start_time >= timeout:
                raise TimeoutError('Waited too long for the port {} on host {} to start accepting '
                                   'connections.'.format(port, host))


class HttpService:
    """Manages a HTTP service instance on localhost. Can start and stop the process.

    Args:
        process_command: Command that will start the service.
            It is a string or a list of strings (like parameters to subprocess.Popen).
            Command strings may contain '{port}' - it will be filled with the provided port.
        port (int): Port on which the service will listen.
            If not provided then the service will run on a random free port.
        env (dict): Environment variables that will be visible for the service process.
            All values of this dictionary must be strings.
            E.g. {'EXAMPLE_VARIABLE_NAME': 'some_example_value'}
            Values may contain '{port}' string - it will be filled with the provided port.
        copy_parent_env (bool): If set to True then environment of the service process will
            contain environment of the parent process updated with `env`.
            If set to False, then only `env` will be set as the service's environment.

    Attributes:
        port (int): Localhost port taken by the service.
        url (int): Address on which the service is available on when it's started.
    """

    def __init__(self, process_command, port=None, env=None, copy_parent_env=True):
        if port is None:
            self.port = port_for.select_random()
        else:
            self.port = port
        self.url = 'http://localhost:{}'.format(self.port)
        self._process_command = self._format_process_command(process_command, self.port)
        self._service_env = self._format_process_env(copy_parent_env, env, self.port)
        self._service_proc = None

    def start(self, timeout=5.0):
        """Starts service process and waits for it to start accepting connections.

        Args:
            timeout (float): How long to wait (in seconds) before raising an error.

        Raises:
            TimeoutError: If the service process didn't start in time.
        """
        self._service_proc = subprocess.Popen(self._process_command, env=self._service_env)
        atexit.register(self.stop)

        try:
            wait_for_port(self.port, timeout=timeout)
        except Exception:
            logging.exception("Service '%s' didn't start", self._process_command)
            self.stop()
            raise

    def stop(self, timeout=5.0):
        """Signals the service process to close and waits for it.

        Args:
            timeout (float): How long to wait (in seconds) before raising an error.

        Raises:
            `subprocess.TimeoutExpired`: If the service process didn't start in time.
        """
        atexit.unregister(self.stop)
        # Sending SIGINT (ctrl+C) because Python handles it by default.
        # Terminate could also be sent, but if the service process spawned
        # another process, then it would need to intercept SIGTERM if we'd
        # want to have a multiprocess coverage report.
        self._service_proc.send_signal(signal.SIGINT)
        self._service_proc.wait(timeout)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    @staticmethod
    def _format_process_command(command, port):
        if not isinstance(command, str):
            return [part.format(port=port) if ('{port}' in part) else part for part in command]
        else:
            return command

    @staticmethod
    def _format_process_env(copy_parent_env, env, port):
        env = env or {}

        formatted_env = env.copy()
        for key, value in env.items():
            if '{port}' in value:
                formatted_env[key] = value.format(port=port)

        if copy_parent_env:
            final_env = os.environ.copy()
            final_env.update(formatted_env)
            return final_env
        else:
            return formatted_env


class ServiceGroup:
    """Manages a group of service processes.
    Can be used to concurrently start or stop more than one service.

    Args:
        *service_processes (list[`HttpService`]): A list of not yet started HTTP services.
    """

    def __init__(self, *service_processes):
        self._services = service_processes
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=7)

    def start(self, timeout=5.0):
        """Starts all the service processes and waits them to start accepting connections.

        Args:
            timeout (float): How long (in seconds) to wait before raising an error.

        Raises:
            TimeoutError: If all of the services didn't start in time.
        """
        start_futures = [self._executor.submit(service.start) for service in self._services]
        results = concurrent.futures.wait(start_futures, timeout=timeout)
        if results.not_done:
            raise TimeoutError('Not all processes started in time.')

    def stop(self, timeout=5.0):
        """
        Stops all the service processes. Waits for full stop.
        Args:
            timeout (float): How long (in seconds) to wait before raising an error.

        Raises:
            TimeoutError: If all of the services didn't stop in time.
        """
        stop_futures = [self._executor.submit(service.stop) for service in self._services]
        results = concurrent.futures.wait(stop_futures, timeout=timeout)
        if results.not_done:
            raise TimeoutError('Not all processes stopped in time.')

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



"""
Abstractions representing aspects of Mountebank.
"""

import collections
import time

import dateutil.parser
import port_for
import requests

from .http_service import HttpService, wait_for_port
from .mb_mgmt import get_mb_command


class MountebankWrapper:
    """A wrapper around the Mountebank API. Meant to be used as a superclass.
    """

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self._imposters_url = 'http://{}:{}/imposters'.format(host, port)

    def add_imposter(self, imposter_cfg):
        """Adds a HTTP service stub (imposter) to Mountebank instance.

        Args:
            imposter_cfg (dict): JSON with Mountebank configuration for an impostor.
                Consult Mountebank documentation.

        Returns:
            `Imposter`: The created service stub.
        """
        resp = requests.post(self._imposters_url, json=imposter_cfg)
        resp.raise_for_status()
        return Imposter(self.port, imposter_cfg['port'], host=self.host)

    def add_imposter_simple(self, port=None, method='GET',  # pylint: disable=too-many-arguments
                            path='/', status_code=200, response=''):
        """Adds an imposter with a single HTTP service stub to Mountebank instance.
        Takes a simplified configuration.

        Args:
            port (int): Port the imposter will listen on. If none, a random port will be selected.
            method (str): HTTP method that the imposter will wait for. GET by default.
            path (str): HTTP path the imposter will wait for. '/' by default.
            status_code (int): HTTP status code the imposter will return. 200 by default.
            response (str): body of the imposters response. Empty string by default.

        Returns:
            `Imposter`: The newly created imposter.
        """
        if port is None:
            port = port_for.select_random()

        return self.add_multi_stub_imposter_simple(
            port,
            [HttpStub(method, path, status_code, response)])

    def add_multi_stub_imposter_simple(self, port, stubs):
        """Adds a Mountebank imposter with multiple HTTP stubs on one port.
        Takes a simplified configuration in comparison to `add_imposter`.

        Args:
            port (int): Port the imposter will listen on.
            stubs (list[`HttpStub`]): HTTP stubs to be created on the port.

        Returns:
            `Imposter`: The newly created imposter.
        """
        imposter_config = {
            'port': port,
            'protocol': 'http',
            'stubs': []
        }

        for stub in stubs:
            stub_json = {
                'responses': [
                    {
                        'is': {
                            'statusCode': stub.status_code,
                            'headers': {
                                'Content-Type': 'application/json'
                            },
                            'body': stub.response
                        }
                    }
                ],
                'predicates': [
                    {
                        'and': [
                            {
                                'equals': {
                                    'path': stub.path,
                                    'method': stub.method,
                                }
                            },
                        ]
                    }
                ]
            }
            imposter_config['stubs'].append(stub_json)
        return self.add_imposter(imposter_config)

    def reset(self):
        """Removes configured imposters (HTTP stubs)."""
        resp = requests.delete(self._imposters_url)
        resp.raise_for_status()

    def start(self):
        """Make sure the process is running and has a clean configuration"""
        raise NotImplementedError()

    def stop(self):
        """Tear down the process if necessary"""
        raise NotImplementedError()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()


class Mountebank(MountebankWrapper):
    """Manages a Mountebank instance. Can start and stop the Mountebank process.

    Args:
        port (int): Port on which Mountebank is listening for imposter configuration commands.
    """
    def __init__(self, port=None):
        process = HttpService(get_mb_command() + ['--mock', '--port', '{port}'], port)
        super().__init__('localhost', process.port)
        self.process = process

    def start(self):
        """Starts the Mountebank process"""
        self.process.start()

    def stop(self):
        """Stops the Mountebank process"""
        self.process.stop()


class ExistingMountebank(MountebankWrapper):
    """Manages existing Mountebank instance. Cannot start or stop the Mountebank process.

    Args:
        host (str): Host on which Mountebank is listening for imposter configuration commands.
        port (int): Port on which Mountebank is listening.
    """

    def start(self):
        """Reset the Mountebank process"""
        wait_for_port(self.port, host=self.host)
        self.reset()

    def stop(self):
        """Reset the Mountebank process"""
        self.reset()


class Imposter:
    """A Mountebank imposter. It can contain stubs of HTTP, HTTPS, TCP or SMTP services.

    Args:
        mountebank_port (int): Mountebank's localhost port.
        port (int): Port for the imposter..

    Attributes:
        url (str): Management URL for the Imposter.
        port (int): Port on localhost taken by this Imposter.
    """

    def __init__(self, mountebank_port, port, host='localhost'):
        self.url = 'http://{}:{}/imposters/{}'.format(host, mountebank_port, port)
        self.port = port

    def requests(self):
        """
        Returns:
            list[`ImposterRequest`]: The requests made on the impostor.
        """
        imposter_json = requests.get(self.url).json()
        imposter_requests = []
        for request in imposter_json.get('requests', []):
            imposter_requests.append(ImposterRequest(
                body=request['body'],
                headers=request['headers'],
                method=request['method'],
                path=request['path'],
                query=request['query'],
                request_from=request['requestFrom'],
                timestamp=dateutil.parser.parse(request['timestamp']),
            ))
        return imposter_requests

    def wait_for_requests(self, count=1, timeout=5.0):
        """Wait until a number of requests arrive to the imposter.

        Args:
            count (int): How many requests to wait for.
            timeout (float): How long to wait for requests.

        Returns:
            list[`ImposterRequest`]: The requests made on the impostor.
        """
        start_time = time.perf_counter()
        while True:
            received_requests = self.requests()
            if len(received_requests) >= count:
                return received_requests
            else:
                time.sleep(0.01)
                if time.perf_counter() - start_time >= timeout:
                    raise TimeoutError('Waited too long for requests on stub.')

    def destroy(self):
        """Deletes this `Imposter` from Mountebank.
        This object cannot be used afterwards.
        """
        requests.delete(self.url)


ImposterRequest = collections.namedtuple(
    'ImposterRequest',
    'body, headers, method, path, query, request_from, timestamp')
ImposterRequest.__doc__ = """Data of a request made on an imposter.

Attributes:
    body (str): Request's body. Can be any valid JSON (this includes raw values)
    headers (dict): HTTP headers.
    method (str): HTTP method.
    path (str): Request's path (e.g. "/bla/1" or "/").
    query (dict): Query arguments (after "?" in a URL).
    request_from (str): Request's source address.
    timestamp (`datetime.datetime`): Time at which the request was made.
"""

HttpStub = collections.namedtuple('HttpStub', ['method', 'path', 'status_code', 'response'])
HttpStub.__doc__ = """A configuration for a simple Mountebank impostor, not including the port.

Attributes:
    method (str): HTTP method to which the stub will answer.
    path (str): Request's path (e.g. "/bla/1" or "/")
    status_code (int): Status code that will be returned by the stub.
    response (str): Stub will send it in response to a request matching other parameters.
"""
